import os
from pathlib import Path

import llvmlite.binding as llvm


def _initialize_llvm():
    # this initializes the per-process LLVM state. It's save to call multiple times.
    # TODO we never call llvm.shutdown(), is this a problem?
    # some parts of the llvm memory are only deallocated once the process exits
    llvm.initialize()
    llvm.initialize_native_target()
    llvm.initialize_native_asmprinter()


def _get_target_machine(fcodemodel="large"):
    target = llvm.Target.from_triple(llvm.get_process_triple())
    try:
        # LLVM raises if features cannot be detected
        features = llvm.get_host_cpu_features().flatten()
    except RuntimeError:
        features = ""

    # large codemodel is necessary for large, ~1000 tree models.
    # for smaller models "default" codemodel would be faster.
    target_machine = target.create_target_machine(
        cpu=llvm.get_host_cpu_name(),
        features=features,
        reloc="pic",
        codemodel=fcodemodel,
    )
    return target_machine


def compile_module_to_asm(module, cache_path=None, fcodemodel="large"):
    _initialize_llvm()

    # Create a target machine representing the host
    target_machine = _get_target_machine(fcodemodel)

    # Create execution engine for our module
    execution_engine = llvm.create_mcjit_compiler(module, target_machine)
    module.data_layout = str(execution_engine.target_data)

    # when caching we dump the executable once the module finished compiling
    # we only ever have one module, hence we can ignore the 'llvm_module' parameter
    # if the module is already cached we load the bytes without any cache-consistency checks
    if cache_path:
        if Path(cache_path).exists():
            execution_engine.set_object_cache(
                getbuffer_func=lambda _: Path(cache_path).read_bytes()
            )
        else:
            execution_engine.set_object_cache(
                notify_func=lambda _, buffer: Path(cache_path).write_bytes(buffer)
            )

    # compile IR to ASM
    execution_engine.finalize_object()
    execution_engine.run_static_constructors()

    if os.environ.get("LLEAVES_PRINT_ASM") == "1":
        print(target_machine.emit_assembly(module))

    return execution_engine



from enum import Enum

ISSUE_ERROR_MSG = "Please file an issue at https://github.com/siboehm/lleaves."


class MissingType(Enum):
    """
    Codes for attribute-values that are treated as missing values when present in a record.
    """

    MNone = 0
    MZero = 1
    MNaN = 2


class DecisionType:
    """
    The different decision types that a node can implement.
    """

    CAT_MASK = 1
    DEFAULT_LEFT_MASK = 2

    def __init__(self, idx):
        if idx not in range(12):
            raise ValueError(
                f"Decision type {idx} not yet tested for. {ISSUE_ERROR_MSG}"
            )
        self.idx = idx

    @property
    def is_categorical(self):
        return bool(self.idx & DecisionType.CAT_MASK)

    @property
    def is_default_left(self):
        assert not self.is_categorical
        return bool(self.idx & DecisionType.DEFAULT_LEFT_MASK)

    @property
    def missing_type(self):
        missing_type = (self.idx >> 2) & 3
        return MissingType(missing_type)

    def __str__(self):
        if self.is_categorical:
            return "=="
        else:
            return "<="



from dataclasses import dataclass, field
from typing import List

from lleaves.compiler.utils import DecisionType


class Node:
    @property
    def is_leaf(self):
        return isinstance(self, LeafNode)


@dataclass
class Tree:
    idx: int
    root_node: Node
    features: list
    class_id: int

    def __str__(self):
        return f"tree_{self.idx}"


@dataclass
class Forest:
    trees: List[Tree]
    features: list
    n_classes: int
    objective_func: str
    objective_func_config: str
    raw_score: bool = False

    @property
    def n_args(self):
        return len(self.features)


@dataclass
class DecisionNode(Node):
    # the threshold in bit-representation if this node is categorical
    cat_threshold: List[int] = field(default=None, init=False)

    # child nodes
    left: Node = field(default=None, init=False)
    right: Node = field(default=None, init=False)

    idx: int
    split_feature: int
    threshold: int
    decision_type: DecisionType
    left_idx: int
    right_idx: int

    def add_children(self, left, right):
        self.left = left
        self.right = right

    def finalize_categorical(self, cat_threshold):
        self.cat_threshold = cat_threshold
        self.threshold = int(self.threshold)

    def validate(self):
        if self.decision_type.is_categorical:
            assert self.cat_threshold is not None
        else:
            assert self.threshold

    def __str__(self):
        return f"node_{self.idx}"


@dataclass
class LeafNode(Node):
    idx: int
    value: float

    def __str__(self):
        return f"leaf_{self.idx}"



import concurrent.futures
import math
import os
from ctypes import CFUNCTYPE, POINTER, c_double, c_int
from pathlib import Path

import llvmlite.binding
import numpy as np

from lleaves import compiler
from lleaves.data_processing import (
    data_to_ndarray,
    extract_model_global_features,
    extract_pandas_traintime_categories,
    ndarray_to_ptr,
)
from lleaves.llvm_binding import compile_module_to_asm

ENTRY_FUNC_TYPE = CFUNCTYPE(
    None,  # return void
    POINTER(c_double),  # pointer to data array
    POINTER(c_double),  # pointer to results array
    c_int,  # start index
    c_int,  # end index
)


class Model:
    """
    The base class of lleaves.
    """

    # machine-targeted compiler & exec engine.
    # We keep this as an object property to protect the compiled binary from being garbage-collected
    _execution_engine = None

    # number of features (=columns)
    _n_feature = None
    # number of classes
    _n_classes = None

    # prediction function, drops GIL on entry
    _c_entry_func = None

    def __init__(self, model_file):
        """
        Initialize the uncompiled model.

        :param model_file: Path to the model.txt.
        """
        self.model_file = model_file
        self.is_compiled = False

        self._pandas_categorical = extract_pandas_traintime_categories(model_file)
        num_attrs = extract_model_global_features(model_file)
        self._n_feature = num_attrs["n_feature"]
        self._n_classes = num_attrs["n_class"]
        self._n_trees = num_attrs["n_trees"]

    def num_feature(self):
        """
        Returns the number of features used by this model.
        """
        return self._n_feature

    def num_model_per_iteration(self):
        """
        Returns the number of models per iteration.

        This is equal to the number of classes for multiclass models, else will be 1.
        """
        return self._n_classes

    def num_trees(self):
        """
        Returns the number of trees in this model.
        """
        return self._n_trees

    def compile(
        self,
        cache=None,
        *,
        raw_score=False,
        fblocksize=34,
        fcodemodel="large",
        finline=True,
    ):
        """
        Generate the LLVM IR for this model and compile it to ASM.

        For most users tweaking the compilation flags (fcodemodel, fblocksize) will be unnecessary as the default
        configuration is already very fast.
        Modifying the flags is useful only if you're trying to squeeze out the last few percent of performance.

        The compile() method is generally not thread-safe.

        :param cache: Path to a cache file. If this path doesn't exist, binary will be dumped at path after compilation.
            If path exists, binary will be loaded and compilation skipped.
            No effort is made to check staleness / consistency.
        :param raw_score: If true, compile the tree to always return raw predictions, without applying
            the objective function. Equivalent to the `raw_score` parameter of LightGBM's Booster.predict().
        :param fblocksize: Trees are cache-blocked into blocks of this size, reducing the icache miss-rate.
            For deep trees or small caches a lower blocksize is better. For single-row predictions cache-blocking
            adds overhead, set `fblocksize=Model.num_trees()` to disable it.
        :param fcodemodel: The LLVM codemodel. Relates to the maximum offsets that may appear in an ASM instruction.
            One of {"small", "large"}.
            The small codemodel will give speedups for most forests, but will segfault when used for compiling
            very large forests.
        :param finline: Whether or not to inline function. Setting this to False will speed-up compilation time
            significantly but will slow down prediction.
        """
        assert 0 < fblocksize
        assert fcodemodel in ("small", "large")

        if cache is None or not Path(cache).exists():
            module = compiler.compile_to_module(
                self.model_file,
                raw_score=raw_score,
                fblocksize=fblocksize,
                finline=finline,
            )
        else:
            # when loading binary from cache we use a dummy empty module
            module = llvmlite.binding.parse_assembly("")

        # keep a reference to the engine to protect it from being garbage-collected
        self._execution_engine = compile_module_to_asm(
            module, cache, fcodemodel=fcodemodel
        )

        # Drops GIL during call, re-acquires it after
        addr = self._execution_engine.get_function_address("forest_root")
        self._c_entry_func = ENTRY_FUNC_TYPE(addr)

        self.is_compiled = True

    def predict(self, data, n_jobs=os.cpu_count()):
        """
        Return predictions for the given data.

        The model needs to be compiled before prediction.

        :param data: Pandas df, numpy 2D array or Python list. Shape should be (n_rows, model.num_feature()).
            2D float64 numpy arrays have the lowest overhead.
        :param n_jobs: Number of threads to use for prediction. Defaults to number of CPUs. For single-row prediction
            this should be set to 1.
        :return: 1D numpy array, dtype float64.
            If multiclass model: 2D numpy array of shape (n_rows, model.num_model_per_iteration())
        """
        if not self.is_compiled:
            raise RuntimeError(
                "Functionality only available after compilation. Run model.compile()."
            )

        # convert all input types to numpy arrays
        data = data_to_ndarray(data, self._pandas_categorical)
        n_predictions = data.shape[0]
        if len(data.shape) != 2 or data.shape[1] != self.num_feature():
            raise ValueError(
                f"Data must be of dimension (N, {self.num_feature()}), is {data.shape}."
            )

        # setup input data and predictions array
        ptr_data = ndarray_to_ptr(data)

        pred_shape = (
            n_predictions if self._n_classes == 1 else (n_predictions, self._n_classes)
        )
        predictions = np.zeros(pred_shape, dtype=np.float64)
        ptr_preds = ndarray_to_ptr(predictions)

        if n_jobs == 1:
            self._c_entry_func(ptr_data, ptr_preds, 0, n_predictions)
        else:
            batchsize = math.ceil(n_predictions / n_jobs)
            with concurrent.futures.ThreadPoolExecutor(max_workers=n_jobs) as executor:
                for i in range(0, n_predictions, batchsize):
                    executor.submit(
                        lambda start_idx: self._c_entry_func(
                            ptr_data,
                            ptr_preds,
                            start_idx,
                            min(start_idx + batchsize, n_predictions),
                        ),
                        i,
                    )
        return predictions



"""
The Scanner is responsible for iterating over the model.txt and transforming it into a more
usable representation.
It doesn't implement any transformations (expect for type casting).
"""


def scan_model_file(file_path, general_info_only=False):
    res = {"trees": []}

    def read_blocks(file_path):
        with open(file_path, "r") as f:
            while True:
                lines = _get_next_block_of_lines(f)
                if lines:
                    yield lines
                else:
                    break

    blocks = read_blocks(file_path)
    # List of blocks we expect:
    # 1* General Information
    # N* Tree, one block for each tree
    # 1* 'end of trees'
    # followed by these ignored blocks:
    # 1* Feature importances
    # 1* Parameters
    # 1* 'end of parameters'
    # 1* 'pandas_categorical:XXXXX'

    general_info_block = next(blocks)
    assert general_info_block[0] == "tree" and general_info_block[1].startswith(
        "version="
    ), f"{file_path} is not a LightGBM model file"
    res["general_info"] = _scan_block(general_info_block, INPUT_SCAN_KEYS)
    if general_info_only:
        return res

    for block in blocks:
        if block[0].startswith("Tree="):
            res["trees"].append(_scan_tree(block))
        else:
            assert block[0] == "end of trees"
            break
    return res


def _scan_tree(lines):
    struct = _scan_block(lines, TREE_SCAN_KEYS)
    return struct


def _get_next_block_of_lines(file):
    # the only function where the position inside the file is advanced
    result = []
    line = file.readline()
    while line == "\n":
        line = file.readline()
    while line != "\n" and line != "":
        result.append(line.strip())
        line = file.readline()
    return result


class ScannedValue:
    def __init__(self, type: type, is_list=False, null_ok=False):
        self.type = type
        self.is_list = is_list
        self.null_ok = null_ok


INPUT_SCAN_KEYS = {
    "max_feature_idx": ScannedValue(int),
    "num_class": ScannedValue(int),
    "num_tree_per_iteration": ScannedValue(int),
    "version": ScannedValue(str),
    "feature_infos": ScannedValue(str, True),
    "objective": ScannedValue(str, True),
}
TREE_SCAN_KEYS = {
    "Tree": ScannedValue(int),
    "num_leaves": ScannedValue(int),
    "num_cat": ScannedValue(int),
    "split_feature": ScannedValue(int, True),
    "threshold": ScannedValue(float, True),
    "decision_type": ScannedValue(int, True),
    "left_child": ScannedValue(int, True),
    "right_child": ScannedValue(int, True),
    "leaf_value": ScannedValue(float, True),
    "cat_threshold": ScannedValue(int, True, True),
    "cat_boundaries": ScannedValue(int, True, True),
}


def _scan_block(lines: list, items_to_scan: dict):
    """
    Scans a block (= list of lines) into a key: value map.
    :param lines: list of lines in the block
    :param items_to_scan: dict with 'key': 'type of value' of keys to scan for
    :return: dict with a key-value pair for each key in items_to_scan. Raises RuntimeError
        if a non-nullable value from items_to_scan wasn't found in the block.
    """
    result_map = {}
    for line in lines:
        # initial line in file
        if line == "tree":
            continue

        scanned_key, scanned_value = line.split("=")
        target_type = items_to_scan.get(scanned_key)
        if target_type is None:
            continue
        if target_type.is_list:
            if scanned_value:
                parsed_value = [target_type.type(x) for x in scanned_value.split(" ")]
            else:
                parsed_value = []
        else:
            parsed_value = target_type.type(scanned_value)
        result_map[scanned_key] = parsed_value

    expected_keys = {k for k, v in items_to_scan.items() if not v.null_ok}
    missing_keys = expected_keys - result_map.keys()
    if missing_keys:
        raise RuntimeError(f"Missing non-nullable keys {missing_keys}")
    return result_map
